import pytest
import pytz

from datetime import datetime
from psi_collector.collector_runner import CollectorRunner
from psi_collector.custom_collector import CustomCollector
from test_utils.mock import MockApi

TEST_CAPTURE_TIME = datetime.fromisoformat('2021-06-01T00:00:00').replace(tzinfo=pytz.utc)
EXPECTED_DATA = [{'dummy': 'data'}, {'empty': ''}, {'number': 0}]


class TestCollector(CustomCollector):
    def __init__(self, capture_time, api, aws_session, http_session):
        super().__init__('test', capture_time, api, aws_session, http_session)

    def _collect(self, s3_path: str = None):
        return EXPECTED_DATA

    def start_date(self, capture_time):
        return capture_time

    def end_date(self, capture_time):
        return capture_time
TestCollector.__test__ = False


class TestCollectorRunner_UnitTests:

    @pytest.mark.unit
    def test_collect(self):
        # Set up mocks
        mock_s3_loader = MockApi()
        mock_s3_loader.on('load', 'test', EXPECTED_DATA, TEST_CAPTURE_TIME).return_val(None)
        mock_timestream_loader = MockApi()
        mock_timestream_loader.on(
            'load',
            data=EXPECTED_DATA,
            capture_time=TEST_CAPTURE_TIME,
            source='test',
            upsert=True,
            ignore_parse_error=True,
            max_batch_size=None,
            min_common_dimensions=None,
            is_multi=None
        ).return_val({'invalid_items': [], 'rejected_records': []})

        # Test collect
        runner = CollectorRunner(
            collectors={'test': TestCollector},
            s3_loader=mock_s3_loader,
            timestream_loader=mock_timestream_loader
        )
        runner.collect({
            'collector_name': 'test',
            'capture_time': TEST_CAPTURE_TIME.isoformat()
        })
        mock_s3_loader.assert_expectations()
        mock_timestream_loader.assert_expectations()

    @pytest.mark.unit
    def test_collect_invalid_items(self):
        # Set up mocks
        mock_s3_loader = MockApi()
        mock_s3_loader.on('load', 'test', EXPECTED_DATA, TEST_CAPTURE_TIME).return_val(None)
        mock_timestream_loader = MockApi()
        mock_timestream_loader.on(
            'load',
            data=EXPECTED_DATA,
            capture_time=TEST_CAPTURE_TIME,
            source='test',
            upsert=True,
            ignore_parse_error=True,
            max_batch_size=None,
            min_common_dimensions=None,
            is_multi=None
        ).return_val({'invalid_items': [1], 'rejected_records': []})

        # Test collect with invalid items
        runner = CollectorRunner(
            collectors={'test': TestCollector},
            s3_loader=mock_s3_loader,
            timestream_loader=mock_timestream_loader
        )

        with pytest.raises(Exception) as e_info:
            runner.collect({
                'collector_name': 'test',
                'capture_time': TEST_CAPTURE_TIME.isoformat()
            })
        assert str(e_info.value).startswith('Failed to write to Timestream with invalid items')
        mock_s3_loader.assert_expectations()
        mock_timestream_loader.assert_expectations()

    @pytest.mark.unit
    def test_collect_rejected_records(self):
        # Set up mocks
        mock_s3_loader = MockApi()
        mock_s3_loader.on('load', 'test', EXPECTED_DATA, TEST_CAPTURE_TIME).return_val(None)
        mock_timestream_loader = MockApi()
        mock_timestream_loader.on(
            'load',
            data=EXPECTED_DATA,
            capture_time=TEST_CAPTURE_TIME,
            source='test',
            upsert=True,
            ignore_parse_error=True,
            max_batch_size=None,
            min_common_dimensions=None,
            is_multi=None
        ).return_val({'invalid_items': [], 'rejected_records': [1]})

        # Test collect with rejected records
        runner = CollectorRunner(
            collectors={'test': TestCollector},
            s3_loader=mock_s3_loader,
            timestream_loader=mock_timestream_loader
        )

        with pytest.raises(Exception) as e_info:
            runner.collect({
                'collector_name': 'test',
                'capture_time': TEST_CAPTURE_TIME.isoformat()
            })
        assert str(e_info.value).startswith('Failed to write to Timestream with rejected records')
        mock_s3_loader.assert_expectations()
        mock_timestream_loader.assert_expectations()

    @pytest.mark.unit
    def test_get_capture_time(self):
        event = {
            'capture_time': '2021-07-31T00:00:00',
            's3_path': 's3://bucket/path/to/object/1627689600000'
        }
        expected_capture_time = datetime.fromisoformat('2021-07-31T00:00:00').replace(tzinfo=pytz.utc)
        runner = CollectorRunner({'test': TestCollector}, MockApi(), MockApi())

        # Test get_capture_time from capture_time
        dt = runner.get_capture_time(event)
        assert dt == expected_capture_time

        # Test get_capture_time from s3_path
        event.pop('capture_time')
        dt = runner.get_capture_time(event)
        assert dt == expected_capture_time

        # Test get_capture_time without capture_time or s3_path
        event.pop('s3_path')
        dt = runner.get_capture_time(event)
        assert dt.tzinfo == pytz.utc

    @pytest.mark.unit
    def test_get_collector(self):
        runner = CollectorRunner({'test': TestCollector}, MockApi(), MockApi())
        expected_capture_time = datetime.now().astimezone(pytz.utc)
        # Test exception raised with a bad collector name
        with pytest.raises(Exception) as e_info:
            runner.get_collector('bad collector name', capture_time=expected_capture_time)
        assert str(e_info.value).startswith('Unrecognized collector')
        
        collector = runner.get_collector(
            'test',
            aws_session=MockApi(),
            capture_time=expected_capture_time
        )
        assert type(collector) == TestCollector



import boto3
import logging
import pytz
import requests
import sys

from datetime import datetime
from psi_collector.custom_collector import CustomCollector
from typing import Any, Dict

class CollectorRunner():
    def __init__(
        self,
        collectors: Dict[str, Any],
        s3_loader: Any = None,
        timestream_loader: Any = None,
        logger: logging.Logger = None
    ):
        '''
        Initializes a CollectorRunner.

        Accepts a collectors dict, mapping collector name to collector class.
        For example:
        {
            "divvy": DivvyCollector,
            "pagerduty": PagerDutyCollector
        }
        
        '''
        assert collectors and isinstance(collectors, dict), 'A collectors dict is required to initialize a CollectorRunner'
        assert s3_loader, 'A S3 loader is required to initialize a CollectorRunner'
        assert timestream_loader, 'A Timestream loader is required to initialize a CollectorRunner'
        assert type(logger) in [logging.Logger, type(None)], f'A logger of type logging.Logger is required to initialize a CollectorRunner. Got {logger}'
        self.__collectors = collectors
        self.__s3_loader = s3_loader
        self.__timestream_loader = timestream_loader

        self.__logger = logger
        if not self.__logger:
            handler = logging.StreamHandler(sys.stdout)
            log = logging.getLogger('collector_runner')
            log.setLevel(logging.INFO)
            handler.setLevel(logging.INFO)
            formatter = logging.Formatter(
                '%(asctime)s [%(levelname)s] %(filename)s:%(lineno)d - %(message)s'
            )
            handler.setFormatter(formatter)
            log.addHandler(handler)
            log.propagate = False
            self.__logger = log

    def collect(self, event: dict):
        '''Initializes and runs a collector.

        Accepts an event dictionary with the following required attributes:
            1. collector_name : the name of the collector to run
            2. s3_path        : the full s3 path to the data to collect from S3

        If writing to S3, the following attributes are required:
            1. write_to_s3    : boolean that must be set to true if writing to S3
            2. bucket_name    : S3 bucket to write to
            3. output_s3_path : S3 path to write to
        
        If writing to Timestream, the following attributes are required:
            1. write_to_timestream     : boolean that must be set to true if writing to Timestream
            2. database_name           : Timestream database to write to
            3. output_timestream_table : Timestream table to write to  
            (optional) is_upsert       : whether or not to upsert Timestream records (default True)
        
        Optionally, capture_time may be set on the event as a millisecond timestamp. If unset,
        capture_time is taken from either the s3_path or the current time.
        '''
        t0 = datetime.now()
        self.__collect(
            collector_name=event.get('collector_name'),
            capture_time=self.get_capture_time(event),
            write_to_s3=event.get('write_to_s3', True),
            input_s3_path=event.get('s3_path'),
            write_to_timestream=event.get('write_to_timestream', True),
            upsert=event.get('is_upsert', True),
            max_batch_size=event.get('max_batch_size'),
            min_common_dimensions=event.get('min_common_dimensions'),
            is_multi=event.get('is_multi'),
            account_id=event.get('account_id')
        )
        self.__logger.info(f'Collection completed in {str(datetime.now() - t0)}')

    def __collect(
        self,
        collector_name: str,
        capture_time: datetime,
        write_to_s3: bool,
        input_s3_path: str,
        write_to_timestream: bool,
        upsert: bool,
        max_batch_size: int,
        min_common_dimensions: int,
        is_multi: bool,
        account_id: str
    ):
        collector = self.get_collector(collector_name, capture_time=capture_time)
        data = collector.collect(s3_path=input_s3_path, account_id=account_id)

        if write_to_s3:
            self.__logger.info('Beginning upload to S3')
            self.__s3_loader.load(collector_name, data, capture_time)

        if write_to_timestream:
            self.__logger.info('Beginning write to Timestream')
            result = self.__timestream_loader.load(
                data=data,
                capture_time=capture_time,
                source=collector_name,
                upsert=upsert,
                ignore_parse_error=True,
                max_batch_size=max_batch_size,
                min_common_dimensions=min_common_dimensions,
                is_multi=is_multi
            )
            if len(result['invalid_items']) > 0:
                raise Exception(f"Failed to write to Timestream with invalid items {result['invalid_items']}")
            if len(result['rejected_records']) > 0:
                raise Exception(f"Failed to write to Timestream with rejected records {result['rejected_records']}")

    def get_capture_time(self, event: dict) -> datetime:
        if 'capture_time' in event and event['capture_time']:
            # fromisoformat returns a naive datetime object that IS NOT adjusted for the local timezone
            return datetime.fromisoformat(event['capture_time']).replace(tzinfo=pytz.utc)
        elif 's3_path' in event and event['s3_path']:
            capture_time_ms = event['s3_path'].split('/')[-1]
            # fromtimestamp returns a naive datetime object that IS adjusted for the local timezone
            return datetime.fromtimestamp(int(capture_time_ms)/1000).astimezone().astimezone(pytz.utc)
        else:
            # now returns a naive datetime object that IS adjusted for the local timezone
            return datetime.now().astimezone().astimezone(pytz.utc)

    def get_collector(
        self,
        collector_name: str,
        api=None,
        aws_session: boto3.Session = None,
        http_session: requests.Session = None,
        capture_time: datetime = None
    ) -> CustomCollector:
        assert capture_time, 'A capture_time is required to initialize a collector'
        cls = self.__collectors.get(collector_name)
        if cls:
            return cls(
                capture_time=capture_time,
                api=api,
                aws_session=aws_session,
                http_session=http_session
            )
        raise Exception(f'Unrecognized collector with name: {collector_name}')



import pytest
import pytz

from datetime import datetime
from psi_collector.custom_collector import CustomCollector
from test_utils.mock import MockApi

TEST_CAPTURE_TIME = datetime.fromisoformat('2021-06-01T00:00:00').replace(tzinfo=pytz.utc)


class TestCollector(CustomCollector):
    def __init__(self, aws_session = None):
        super().__init__('test', TEST_CAPTURE_TIME, None, aws_session, None)

    def _collect(self, s3_path: str = None):
        return [{'dummy': 'data'}, {'empty': ''}, {'number': 0}]

    def start_date(self, capture_time):
        return capture_time

    def end_date(self, capture_time):
        return capture_time
TestCollector.__test__ = False

class TestCustomCollector_UnitTests:

    @pytest.mark.unit
    def test_init_failure(self):
        # Assert exception raised when initializing (can't initialize an abstract class)
        with pytest.raises(Exception) as e_info:
            c = CustomCollector('test', TEST_CAPTURE_TIME, None, None, None)

    @pytest.mark.unit
    def test_collect(self):
        c = TestCollector()
        data = c.collect()
        assert len(data) == 3
        assert data[0] == {'dummy': 'data'}
        assert data[1] == {'empty': 'null'}
        assert data[2] == {'number': 0}

    @pytest.mark.unit
    def test_format_date(self):
        pacific = pytz.timezone("US/Pacific")
        date_with_tz = datetime(2021, 6, 1).replace(tzinfo=pytz.utc).astimezone(pacific)
        date_without_tz = datetime(2021, 6, 1)

        c = TestCollector()
        assert c.format_date(date_without_tz) == '2021-06-01T00:00:00Z'
        assert c.format_date(date_with_tz) == '2021-05-31T17:00:00-07:00'

    @pytest.mark.unit
    def test_parse_date(self):
        # Assert exception raised when bad date string is given
        c = TestCollector()
        with pytest.raises(Exception) as e_info:
            c.parse_date('bad string raise exception')

        # Parse a UTC date
        dt = c.parse_date('2021-06-01T09:08:07Z')
        assert dt.timestamp() == 1622538487.0
        assert dt.tzname() == 'UTC'

        # Parse a non-UTC date
        dt = c.parse_date('2021-06-01T09:08:07-05:00')
        assert dt.timestamp() == 1622556487.0
        assert dt.tzname() == 'UTC-05:00'

    @pytest.mark.unit
    def test_get_s3_object(self):
        # Set expected values
        expected_exception = Exception('Test exception')

        # Set up mock clients
        mock_s3_client = MockApi()
        mock_s3_client.on(
            'get_object', Bucket='bucket', Key='key'
        ).raise_exception(expected_exception)

        mock_session = MockApi()
        mock_session.on('client', 's3').return_val(mock_s3_client)

        # Test exception raised by S3 client
        c = TestCollector(aws_session=mock_session)
        with pytest.raises(Exception) as e_info:
            c._get_s3_object('s3://bucket/key')

    @pytest.mark.unit
    def test_get_parameter(self):
        # Set expected values
        expected_parameter_name = 'test'
        expected_parameter_value = 'parameter value'

        # Set up mock clients
        mock_ssm_client = MockApi()
        mock_ssm_client.on('get_parameter', Name=expected_parameter_name, WithDecryption=True).return_val({
            'Parameter': {'Value': expected_parameter_value}
        })

        mock_session = MockApi()
        mock_session.on('client', 'ssm').return_val(mock_ssm_client)

        c = TestCollector(aws_session=mock_session)
        value = c._get_parameter('test')
        assert value == expected_parameter_value

    @pytest.mark.unit
    def test_get_parameter_no_parameter_value(self):
        # Set expected values
        expected_parameter_name = 'test'
        expected_parameter_value = 'parameter value'

        # Set up mock clients
        mock_ssm_client = MockApi()
        mock_ssm_client.on(
            'get_parameter', Name=expected_parameter_name
        ).return_val({})

        mock_session = MockApi()
        mock_session.on('client', 'ssm').return_val(mock_ssm_client)

        # Test exception raised when no parameter value in client response
        c = TestCollector(aws_session=mock_session)
        with pytest.raises(Exception) as e_info:
            value = c._get_parameter('test')

    @pytest.mark.unit
    def test_repr(self):
        c = TestCollector()
        assert str(c) == '{"name": "test", "start_date": "2021-06-01T00:00:00Z", "end_date": "2021-06-01T00:00:00Z"}'
